#	include	"Standard.h"
#	include	"mygl.h"
#	include	"render.h"
#	include	"zbuffer.h"

//Vector4		LightSource	=	{	1.0, 0.7, -0.2, 0.0	};
//Vector4		LightSource	=	{	1.0, 1.0, 1.0, 0.0	};
Vector4		LightSource	=	{	0.0, 0.0, 1.0, 0.0	};

Matrix4x4	ViewerMatrix	=
	{
	{	1.0, 0, 0, 0.0	},
	{	0, 1.0, 0, 0.0	},
	{	0, 0, 1.0, 0.0	},
	{	0, 0, 0, 1.0	}
	};

Matrix4x4	PerspectiveMatrix	=
	{
	{	1.0, 0, 0.0, 0	},
	{	0, 1.0, 0.0, 0.0	},
	{	0, 0, 1.0, 0.0	},
	{	0, 0, 0.0, 1.0	}
	};

Matrix4x4	ViewAndPerspectiveMatrix;

// hardware primitive: Matrix4x4 * Vector4

void
Mul(Matrix4x4	m, Vector4	v, Vector4	out)
	{
	Vector4	t;
	for(int	i=0;i<4;i++)
		{
		t[i] = FloatZero;
		for(int	j=0;j<4;j++)
			{
			t[i] += m[i][j] * v[j];
			}
		}
	// extra copy is here just incase v == out
	for(int	k=0;k<4;k++)
		out[k] = t[k];
	}

// hardware primitive: stream a list of polygons through and transform their vertexs by a matrix
void
TransformPolygons(
	Array<Polygon>	*pin,
	Matrix4x4		mView,
	Matrix4x4		mViewAndPerspective,
	Array<Polygon>	*pout)
	{
	int	Index = 0;

	for(Index = 0; Index < pin->GetSize(); Index++)
		{
		Vector4	v, v1;
		Polygon	p = pin->ElementAt(Index);

		VertexToVector4(p.v1, v); Mul(mViewAndPerspective, v, v1); Vector4ToVertex(v1, p.v1);
		VertexToVector4(p.v2, v); Mul(mViewAndPerspective, v, v1); Vector4ToVertex(v1, p.v2);
		VertexToVector4(p.v3, v); Mul(mViewAndPerspective, v, v1); Vector4ToVertex(v1, p.v3);
		Mul(mView,p.n,p.n);	// transform normal
		pout->ElementAti(Index) = p;
		}

	}

bool
LightPolygons(
	Array<Polygon>	*thePolygonList,
	Vector4			LightSource)
	{
	int	Index = 0;
	for(Index = 0; Index < thePolygonList->GetSize(); Index++)
		{
		// compute dot product
		Polygon	p = thePolygonList->ElementAt(Index);
		Float	c = p.n[0] * LightSource[0] + p.n[1] * LightSource[1] + p.n[2] * LightSource[2];

		c = (c + 1.0)/2.0;
		if ( c < 0 )	c = 0;
		if ( c > 1.0 )	c = 1.0;
		thePolygonList->ElementAt(Index).v1.color.r = (byte) (thePolygonList->ElementAt(Index).v1.color.r * c);
		thePolygonList->ElementAt(Index).v1.color.g = (byte) (thePolygonList->ElementAt(Index).v1.color.g * c);
		thePolygonList->ElementAt(Index).v1.color.b = (byte) (thePolygonList->ElementAt(Index).v1.color.b * c);
		}
	return(true);
	}


// project polygons... Similar to Transform except for additional step
void
ProjectPolygons(
	Array<Polygon>	*pin,
	Float			theScreenWidth,
	Float			theScreenHeight,
	Array<Polygon>	*pout)
	{
	int	Index = 0;

	for(Index = 0; Index < pin->GetSize(); Index++)
		{
		Vector4	v1;
		Polygon	p = pin->ElementAt(Index);

		VertexToVector4(p.v1, v1);

		p.v1.x = (((v1[0]) + 1.0) / 2.0) * theScreenWidth;
		p.v1.y = (((-v1[1]) + 1.0) / 2.0) * theScreenHeight;
		p.v1.z = (((-v1[2]) + 1.0) / 2.0) * 256;
		if (p.v1.z > 255) p.v1.z = 255; if (p.v1.z < 0 ) p.v1.z = 0;

		VertexToVector4(p.v2, v1);
		p.v2.x = (((v1[0]) + 1.0) / 2.0) * theScreenWidth;
		p.v2.y = (((-v1[1]) + 1.0) / 2.0) * theScreenHeight;
		p.v2.z = (((-v1[2]) + 1.0) / 2.0) * 256;
		if (p.v2.z > 255) p.v2.z = 255; if (p.v2.z < 0 ) p.v2.z = 0;

		VertexToVector4(p.v3, v1);
		p.v3.x = (((v1[0]) + 1.0) / 2.0) * theScreenWidth;
		p.v3.y = (((-v1[1]) + 1.0) / 2.0) * theScreenHeight;
		p.v3.z = (((-v1[2]) + 1.0) / 2.0) * 256;
		if (p.v3.z > 255) p.v3.z = 255; if (p.v3.z < 0 ) p.v3.z = 0;


		
		pout->ElementAti(Index) = p;
		}

	}

// Bresenham's algorithm, but simplified for drawing a 1 D line and updating a Z-buffer
bool
Rasterize(
	Float	y,
	Float	StartX,
	Float	StartZ,
	Float	EndX,
	Float	EndZ,
	Color	c,
	ZBuffer	*theZBuffer)
	{
	int		iy, iStartX, iStartZ, iEndX, iEndZ;

	if ( StartX > EndX )
		{
		SwapFloats(StartX, EndX);
		SwapFloats(StartZ, EndZ);
		}
	// helps with some round-off errors
	//StartX -= 0.5;
	//EndX += 0.5;

	iy = (int)floor(y + 0.5);
	iStartX = (int) floor(StartX + 0.5);
	iStartZ = (int) floor(StartZ + 0.5);
	iEndX = (int) floor(EndX + 0.5);
	iEndZ = (int) floor(EndZ + 0.5);

	int			dz = iEndZ - iStartZ;
	int			dx = iEndX - iStartX;
	int			StepX, StepZ;
	if ( dz < 0 ) { dz = -1*dz;	StepZ = -1; } else { StepZ = 1; }
	StepX = 1;

	dz = dz << 1;
	dx = dx << 1;
	
	int	Fraction;
	if ( dz > dx )	Fraction = dx - (dz >> 1);	else	Fraction = dz - (dx >> 1);

	int	PixelStart = iy*theZBuffer->Width + iStartX;

	while ( iStartX <= iEndX && iStartX < theZBuffer->Width)
		{
		c.z = iStartZ;
		if ( iStartX >= 0 )	theZBuffer->BlendPixel(PixelStart, c);

		iStartX += StepX;
		PixelStart += StepX;
		Fraction += dz;
		if ( dx )
		while( Fraction >= 0 )
			{
			Fraction -= dx;
			iStartZ += StepZ;
			}
		}
	return ( true );
	}


// This renders a triangle (half of one).  It is based upon the essence of Bresenham's algorithm
// but extended to support 3D.
bool
Rasterize(
	Vertex	bottom,
	Vertex	middle,
	Vertex	top,
	Color	c,
	ZBuffer	*theZBuffer )
	{
	int		iStart[2][3];
	int		iEnd[2][3];
	int		Step[2][3];
	int		Fraction[2][2];
	int		ds[2][3];
	int		i,j;

	iStart[0][0] = iStart[1][0] = (int) floor(bottom.x + 0.5);
	iStart[0][1] = iStart[1][1] = (int) floor(bottom.z + 0.5);
	iStart[0][2] = iStart[1][2] = (int) floor(bottom.y + 0.5);

	iEnd[0][0] = (int) floor(middle.x + 0.5);
	iEnd[0][1] = (int) floor(middle.z + 0.5);
	iEnd[0][2] = (int) floor(middle.y + 0.5);

	iEnd[1][0] = (int) floor(top.x + 0.5);
	iEnd[1][1] = (int) floor(top.z + 0.5);
	iEnd[1][2] = (int) floor(top.y + 0.5);

	for(i=0;i<2;i++)
	for(j=0;j<3;j++)
		ds[i][j] = iEnd[i][j] - iStart[i][j];

	for(i=0;i<2;i++)
	for(j=0;j<3;j++)
		if ( ds[i][j] < 0 )	{ ds[i][j] = -1*ds[i][j];	Step[i][j] = -1; } else { Step[i][j] = 1; }
	
	for(i=0;i<2;i++)
	for(j=0;j<3;j++)
		ds[i][j] = ds[i][j] << 1;

	for(i=0;i<2;i++)
	for(j=0;j<2;j++)
		if ( ds[i][j] > ds[i][2] )	Fraction[i][j] = ds[i][2] - (ds[i][j] >> 1);
		else						Fraction[i][j] = ds[i][j] - (ds[i][2] >> 1);

	if ( iStart[0][2] >= theZBuffer->Height  && iEnd[0][2] >= theZBuffer->Height )	return (false);
	if ( iStart[0][2] < 0 && iEnd[0][2] < 0 )										return (false);

//	printf("Rasterizing: (%d, %d, %d) -> (%d, %d, %d) & (%d, %d, %d) -> (%d, %d, %d)\n",
//		iStart[0][0], iStart[0][2], iStart[0][1],
//		iEnd[0][0], iEnd[0][2], iEnd[0][1],
//		iStart[1][0], iStart[1][2], iStart[1][1],
//		iEnd[1][0], iEnd[1][2], iEnd[1][1]);
		
	while ( iStart[0][2] != iEnd[0][2] )
		{
//		printf("scanline: %d: (%d,%d) -> (%d, %d)\n", iStart[0][2],
//			iStart[0][0],iStart[0][1],
//			iStart[1][0],iStart[1][1]);
		if ( iStart[0][2] < theZBuffer->Height && iStart[0][2] > 0 )
			Rasterize(iStart[0][2], iStart[0][0], iStart[0][1], iStart[1][0], iStart[1][1], c, theZBuffer);

		for(i=0;i<2;i++)
			{
			iStart[i][2] += Step[i][2];
			for(j=0;j<2;j++)
				Fraction[i][j] += ds[i][j];
			for(j=0;j<2;j++)
				while( Fraction[i][j] >= 0 )
					{
					Fraction[i][j] -= ds[i][2];
					iStart[i][j] += Step[i][j];
					}
			}
		}
	if ( iStart[0][2] < theZBuffer->Height && iStart[0][2] > 0 )
		Rasterize(iStart[0][2], iStart[0][0], iStart[0][1], iStart[1][0], iStart[1][1], c, theZBuffer);
	return ( true );
	}

bool
Rasterize(
	Polygon			p,
	ZBuffer			*theZBuffer)
	{
	Vertex	top, middle, bottom;

	// sort the vertexes of p according to their y components
	top = p.v1;
	middle = p.v2;
	bottom = p.v3;

	// simple bubble sort
	if ( top.y < middle.y )		SwapVertexs(top, middle);
	if ( middle.y < bottom.y )	SwapVertexs(middle, bottom);
	if ( top.y < middle.y )		SwapVertexs(top, middle);


	Rasterize(bottom, middle, top, p.v1.color, theZBuffer);
	Rasterize(top, middle, bottom, p.v1.color, theZBuffer);
	return ( true );
	}

bool
Rasterize(
	Array<Polygon>	*thePolygonList,
	ZBuffer			*theZBuffer)
	{
	for(int	Index = 0;Index < thePolygonList->GetSize(); Index++)
		Rasterize(thePolygonList->ElementAt(Index), theZBuffer);
	return(true);
	}

///////////////////////////////////////////////////////////////////////// Main GPU pipeline
//

bool
Render(
	Array<Polygon>	*thePolygonList,
	Vector4			LightSource,
	ZBuffer			*theZBuffer)
	{
	// transform to viewer
	TransformPolygons(thePolygonList, ViewerMatrix, ViewAndPerspectiveMatrix, thePolygonList);

	// clip in the Z direction
	// skip for now, clip in the z-buffer

	// light vertexs
	LightPolygons(thePolygonList, LightSource);

	// transform to viewer
	ProjectPolygons(thePolygonList, theZBuffer->Width, theZBuffer->Height, thePolygonList);

	// clip to the screen
	// skip for now and clip in the screen buffer

	// rasterize to Z-buffer
	Rasterize(thePolygonList, theZBuffer);

	return ( true );
	}



/////////////////////////////////////////////////////////////////////// CPU host functions

// some nifty functons to make some rotation matrixes
void
BuildRx(Float	theta, Matrix4x4	m)
	{
	m[0][0] = 1.0;	m[0][1] = 0;			m[0][2] = 0;			 m[0][3] = 0;
	m[1][0] = 0;	m[1][1] = cos(theta);	m[1][2] = -1*sin(theta); m[1][3] = 0;
	m[2][0] = 0;	m[2][1] = sin(theta);	m[2][2] = cos(theta);	 m[2][3] = 0;
	m[3][0] = 0;	m[3][1] = 0;			m[3][2] = 0;			 m[3][3] = 1.0;
	}

void
BuildRy(Float	theta, Matrix4x4	m)
	{
	m[0][0] = cos(theta);	m[0][1] = 0;	m[0][2] = sin(theta);	m[0][3] = 0;
	m[1][0] = 0;	m[1][1] = 1;			m[1][2] = 0;			m[1][3] = 0;
	m[2][0] = -1*sin(theta);m[2][1] = 0;	m[2][2] = cos(theta);	m[2][3] = 0;
	m[3][0] = 0;	m[3][1] = 0;			m[3][2] = 0;			m[3][3] = 1;
	}

void
BuildRz(Float	theta, Matrix4x4	m)
	{
	m[0][0] = cos(theta);	m[0][1] = -1*sin(theta);	m[0][2] = 0;	m[0][3] = 0;
	m[1][0] = sin(theta);	m[1][1] = cos(theta);		m[1][2] = 0;	m[1][3] = 0;
	m[2][0] = 0;			m[2][1] = 0;				m[2][2] = 1;	m[2][3] = 0;
	m[3][0] = 0;			m[3][1] = 0;				m[3][2] = 0;	m[3][3] = 1;
	}

void
Mul(Matrix4x4	m1, Matrix4x4	m2, Matrix4x4	o)
	{
	Matrix4x4	t;
	int			i,j,k;

	for(i = 0; i < 4; i++ ) {
		for(j = 0; j < 4; j ++ ) {
			t[i][j] = 0;
			for(k = 0;k < 4; k++ ) {
				t[i][j] += m1[i][k] * m2[k][j];
			}
		}
	}

	for(i = 0; i < 4; i++ )
		for(j = 0; j < 4; j ++ )
			o[i][j] = t[i][j];
	}


// useful for debugging rasterization
bool
RandomlyLightPolygons(
	Array<Polygon>	*thePolygonList)
	{
	// this is a random hack for test
	for(int	i=0; i < thePolygonList->GetSize(); i++)
		{
		thePolygonList->ElementAt(i).v1.color.r = rand()%255;
		thePolygonList->ElementAt(i).v1.color.g = rand()%255;
		thePolygonList->ElementAt(i).v1.color.b = rand()%255;
		}
	return(true);
	}

// normals of each polygon are the simplest lighting model
void
NormalVector(
	Polygon	p,
	Vector4	v)
	{
	Vector4	v1, v2;
	
	v1[0] = p.v2.x - p.v1.x;
	v1[1] = p.v2.y - p.v1.y;
	v1[2] = p.v2.z - p.v1.z;
	v1[3] = 0;

	v2[0] = p.v3.x - p.v1.x;
	v2[1] = p.v3.y - p.v1.y;
	v2[2] = p.v3.z - p.v1.z;
	v2[3] = 0;

	v[0] = v1[1] * v2[2] - v1[2] * v2[1];
	v[1] = v1[2] * v2[0] - v1[0] * v2[2];
	v[2] = v1[0] * v2[1] - v1[1] * v2[0];
	v[3] = 0;

	Float	l = sqrt( v[0] * v[0] + v[1] * v[1] + v[2] * v[2] );

	v[0] = v[0] / l;	
	v[1] = v[1] / l;	
	v[2] = v[2] / l;
	}

bool
ComputePolygonNormals(
	Array<Polygon>	*thePolygonList,
	Vector4			NormalizedLightSource)
	{
	Float	l = sqrt( LightSource[0] * LightSource[0] + LightSource[1] * LightSource[1] + LightSource[2] * LightSource[2] );

	NormalizedLightSource[0] = LightSource[0] / l;
	NormalizedLightSource[1] = LightSource[1] / l;
	NormalizedLightSource[2] = LightSource[2] / l;
	NormalizedLightSource[3] = FloatZero;

	for(int	i=0; i < thePolygonList->GetSize(); i++)
		{
		Polygon	p;

		p = thePolygonList->ElementAt(i);
		NormalVector(p, p.n);
		thePolygonList->ElementAt(i) = p;
		}

	return(true);
	}

bool
MakeFrustumMatrix(
	Matrix4x4	m,
	double		l,
	double		r,
	double		b,
	double		t,
	double		n,
	double		f)
	{
	m[0][0] = 2*n*(r-l);	m[0][1] = 0;	m[0][2] = (r+l)/(r-l);	m[0][3] = 0;
	m[1][0] = 0;		m[1][1] = 2*n*(t-b);	m[1][2] = (t+b)/(t-b);	m[1][3] = 0;
	m[2][0] = 0;		m[2][1] = 0;	m[2][2] = -1*(f+n)/(f-n);	m[2][3] = -2*f*n/(f-n);
	m[3][0] = 0;		m[3][1] = 0;	m[3][2] = -1;				m[3][3] = 0;
	return ( true );	
	}


bool
MakeScaleMatrix(
	Matrix4x4	m,
	double		factor)
	{
	m[0][0]	=	factor;	m[0][1] = 0;	m[0][2] = 0;	m[0][3] = 0;
	m[1][0]	=	0;	m[1][1] = factor;	m[1][2] = 0;	m[1][3] = 0;
	m[2][0]	=	0;	m[2][1] = 0;	m[2][2] = factor;	m[2][3] = 0;
	m[3][0]	=	0;	m[3][1] = 0;	m[3][2] = 0;	m[3][3] = factor;
	return(true);
	}

int
main(
	int		argc,
	char	*argv[])
	{
	if ( argc < 3 )
		{
		printf("usage: render <infile.off> <outfile.pbm> [rx] [ry] [rz] [scale]\n");
		return(1);
		}

	Array<Polygon>	*PolygonList = ReadOffFile(argv[1]);

	if ( ! PolygonList )
		{
		printf("error reading input file: %s\n", argv[1]);
		return(2);
		}

	Vector4	LightSource;

	ComputePolygonNormals(PolygonList, LightSource);

	//RandomlyLightPolygons(PolygonList);

	// possibly modify the viewer matrix depending upon parameters
	Matrix4x4	m;
	if ( argc > 3 )
		{
		BuildRx(atof(argv[3]), m);
		Mul(ViewerMatrix, m, ViewerMatrix);
		}
	if ( argc > 4 )
		{
		BuildRy(atof(argv[4]), m);
		Mul(ViewerMatrix, m, ViewerMatrix);
		}
	if ( argc > 5 )
		{
		BuildRz(atof(argv[5]), m);
		Mul(ViewerMatrix, m, ViewerMatrix);
		}
	if ( argc > 6 )
		{
		MakeScaleMatrix(m, atof(argv[6]));
		Mul(ViewerMatrix, m, ViewerMatrix);
		}

	MakeFrustumMatrix(PerspectiveMatrix, -1.0, 1.0, 1.0, -1.0, -0.1, 1.0);

	Mul(ViewerMatrix, PerspectiveMatrix, ViewAndPerspectiveMatrix);

	ZBuffer			theZBuffer(640,480);
	theZBuffer.Reset();

	Render(PolygonList, LightSource, &theZBuffer);

	theZBuffer.Dump(argv[2]);
	return(0);
	}

